package ru.dbotthepony.mc.otm.menu.data;

import net.minecraft.client.Minecraft;
import net.minecraft.network.FriendlyByteBuf;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.fmllegacy.network.NetworkEvent;
import net.minecraftforge.fmllegacy.network.PacketDistributor;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;
import java.util.function.Supplier;

public class NetworkedItemView {
	public record ViewedItem(int id, ItemStack stack, UUID id_upstream) {
		public ViewedItem(int id, ItemStack stack) {
			this(id, stack, null);
		}
	}

	public record ClearPacket(UUID id, boolean remove) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeLong(id.getMostSignificantBits());
			buffer.writeLong(id.getLeastSignificantBits());
			buffer.writeBoolean(remove);
		}

		public static ClearPacket read(FriendlyByteBuf buffer) {
			return new ClearPacket(new UUID(buffer.readLong(), buffer.readLong()), buffer.readBoolean());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = TRACKERS_REMOTE.get(id);

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				if (remove) {
					get.remove();
				} else {
					get.clear();
				}
			});
		}
	}

	public record CreatedPacket(UUID id) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeLong(id.getMostSignificantBits());
			buffer.writeLong(id.getLeastSignificantBits());
		}

		public static CreatedPacket read(FriendlyByteBuf buffer) {
			return new CreatedPacket(new UUID(buffer.readLong(), buffer.readLong()));
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				new NetworkedItemView(Minecraft.getInstance().player, id, true);
			});
		}
	}

	public record StackAddPacket(UUID id, int stack_id, ItemStack stack) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeLong(id.getMostSignificantBits());
			buffer.writeLong(id.getLeastSignificantBits());
			buffer.writeInt(stack_id);
			buffer.writeRegistryId(stack.getItem());
			buffer.writeInt(stack.getCount());
			buffer.writeBoolean(stack.getTag() != null);

			if (stack.getTag() != null)
				buffer.writeNbt(stack.getShareTag());
		}

		public static StackAddPacket read(FriendlyByteBuf buffer) {
			var id = new UUID(buffer.readLong(), buffer.readLong());
			var stack = buffer.readInt();
			var item = buffer.readRegistryIdSafe(Item.class);
			var count = buffer.readInt();

			var state = new ItemStack(item, count);

			if (buffer.readBoolean())
				state.readShareTag(buffer.readNbt());

			return new StackAddPacket(id, stack, state);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = TRACKERS_REMOTE.get(id);

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				if (get.state.containsKey(stack_id)) {
					throw new IllegalStateException("Item tracker " + id + " already has stack with id of " + stack_id);
				}

				get.state.put(stack_id, new ViewedItem(stack_id, stack));
				get.clearCache();
			});
		}
	}

	public record StackChangePacket(UUID id, int stack_id, int new_count) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeLong(id.getMostSignificantBits());
			buffer.writeLong(id.getLeastSignificantBits());
			buffer.writeInt(stack_id);
			buffer.writeInt(new_count);
		}

		public static StackChangePacket read(FriendlyByteBuf buffer) {
			var id = new UUID(buffer.readLong(), buffer.readLong());
			var stack_id = buffer.readInt();
			var new_count = buffer.readInt();

			return new StackChangePacket(id, stack_id, new_count);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = TRACKERS_REMOTE.get(id);

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				var get_state = get.state.get(stack_id);

				if (get_state == null) {
					throw new IllegalStateException("Item tracker " + id + " has no stack with id of " + stack_id);
				}

				get_state.stack.setCount(new_count);
				get.clearCache();
			});
		}
	}

	public record StackRemovePacket(UUID id, int stack_id) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeLong(id.getMostSignificantBits());
			buffer.writeLong(id.getLeastSignificantBits());
			buffer.writeInt(stack_id);
		}

		public static StackRemovePacket read(FriendlyByteBuf buffer) {
			var id = new UUID(buffer.readLong(), buffer.readLong());
			var stack_id = buffer.readInt();

			return new StackRemovePacket(id, stack_id);
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var get = TRACKERS_REMOTE.get(id);

				if (get == null) {
					throw new IllegalStateException("No such item tracker with id " + id);
				}

				var get_state = get.state.remove(stack_id);

				if (get_state == null) {
					throw new IllegalStateException("Item tracker " + id + " has no stack with id of " + stack_id);
				}

				get.clearCache();
			});
		}
	}

	public final Player ply;
	public final UUID id;
	public final boolean remote;
	protected int next_stack_id = 0;

	protected final HashMap<Integer, ViewedItem> state = new HashMap<>();
	protected final HashMap<UUID, ViewedItem> upstream_state = new HashMap<>();
	protected final ArrayList<Object> backlog = new ArrayList<>();

	// normally, filled only on server
	private static final HashMap<UUID, NetworkedItemView> TRACKERS_LOCAL = new HashMap<>();

	// normally, filled only on client
	private static final HashMap<UUID, NetworkedItemView> TRACKERS_REMOTE = new HashMap<>();

	public static NetworkedItemView getLocal(UUID id) {
		return TRACKERS_LOCAL.get(id);
	}

	public static NetworkedItemView getRemote(UUID id) {
		return TRACKERS_REMOTE.get(id);
	}

	public NetworkedItemView(Player ply) {
		this(ply, UUID.randomUUID());
	}

	public NetworkedItemView(Player ply, UUID id) {
		this(ply, id, false);
	}

	public NetworkedItemView(Player ply, UUID id, boolean remote) {
		this.remote = remote;
		this.ply = ply;
		this.id = id;

		if (remote) {
			TRACKERS_REMOTE.put(this.id, this);
		} else {
			TRACKERS_LOCAL.put(this.id, this);
			MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with(() -> (ServerPlayer) ply), new CreatedPacket(this.id));
		}
	}

	private List<ItemStack> view_cache;

	public void clearCache() {
		view_cache = null;
	}

	public List<ItemStack> getItems() {
		if (view_cache != null)
			return view_cache;

		var list = new ArrayList<ItemStack>(state.size());

		for (var state : this.state.values())
			list.add(state.stack);

		return view_cache = list;
	}

	public void addItem(ItemStack stack, UUID id_upstream) {
		if (upstream_state.containsKey(id_upstream))
			throw new IllegalStateException("Already tracking ItemStack with upstream id " + id_upstream + "!");

		var state = new ViewedItem(next_stack_id++, stack, id_upstream);
		this.state.put(state.id, state);
		this.upstream_state.put(id_upstream, state);

		if (!remote) {
			backlog.add(new StackAddPacket(id, state.id, stack));
		}

		clearCache();
	}

	public void changeItem(UUID id_upstream, int new_count) {
		var get = upstream_state.get(id_upstream);

		if (get == null)
			throw new IllegalStateException("Unknown ItemStack with upstream id " + id_upstream + "!");

		get.stack.setCount(new_count);

		if (!remote) {
			backlog.add(new StackChangePacket(id, get.id, new_count));
		}

		clearCache();
	}

	public void removeItem(UUID id_upstream) {
		var get = upstream_state.get(id_upstream);

		if (get == null)
			throw new IllegalStateException("Unknown ItemStack with upstream id " + id_upstream + "!");

		upstream_state.remove(id_upstream);
		state.remove(get.id);

		if (!remote) {
			backlog.add(new StackRemovePacket(id, get.id));
		}

		clearCache();
	}

	public void clear() {
		clearCache();
		upstream_state.clear();
		state.clear();

		if (!remote) {
			backlog.clear();
			backlog.add(new ClearPacket(id, false));
		}
	}

	public void remove() {
		if (!remote) {
			TRACKERS_LOCAL.remove(id);
			backlog.clear();
			backlog.add(new ClearPacket(id, true));
		}
	}

	public void network() {
		if (remote)
			throw new IllegalStateException("Not a server");

		var consumer = PacketDistributor.PLAYER.with(() -> (ServerPlayer) ply);

		for (var packet : backlog) {
			MatteryNetworking.CHANNEL.send(consumer, packet);
		}

		backlog.clear();
	}
}
